class FileParser:
    """ FileParser is a class used to parse both lines and words from a file. I will also sterilize the data """

    # TODO: When cleaning lines. Check that it's not a whitelisted use of a period

    def __init__(self, file_name):
        """ We read a file named :file_name: and put all of the words and lines into a variable

        :param file_name:       The name of the file we want to parse
        """

        file_reader = open(file_name, "r", encoding="utf8")

        self.words = []
        self.lines = file_reader.read().splitlines()

        # Goes through all the lines, cleans them, then splits it up into words and adds them to the words variable
        for line in self.lines:
            self.words += line.split()

    @staticmethod
    def clean_line(line):
        """ We clean a line. At the moment we just add the [NONE] token at the end of sentences.

        :param line:    The line we want to clean
        :return:        The cleaned line
        """

        return line.replace('.', ' [NONE]')



import random


class HistogramWithCount:
    """ Histogram is a class to store types and tokens

     We store all our data in an array of arrays. The internal arrays we will be calling nodes. Each node represents
     a respected number of word occurrences. The node has three values at the indexes 0, 1.

     0: The number of occurrences
     1: A list of words that occur (0) amount of times
     """

    def __init__(self):
        """ Initialize the data """
        self.nodes = []
        self.word_count = 0

    def __len__(self):
        """ Gets the amount of words in the data set. Not types! """
        return self.word_count

    def __str__(self):
        """ THe string representation of the nodes """
        return str(self.nodes)

    def update_word(self, word):
        """ If the word is already in the histogram we will update the number of times the word occurs. If not then
        we insert it into our nodes

        :param word:     The word we want update in the data set
        """

        self.word_count += 1
        length = len(self.nodes)

        # If we have no data. Let's just create a new node.
        if length < 1:
            self.nodes.append([1, [word]])
            return

        for i in range(length):
            current_node = self.nodes[i]
            words = current_node[1]

            # If our word is in the current node. Remove it then check more things.
            if word in words:
                words.remove(word)

                # If there are no following nodes. We are the greatest node and can just add a new node
                if i > length - 2:
                    self.nodes.append([current_node[0] + 1, [word]])
                else:
                    next_node = self.nodes[i + 1]

                    # If the next nodes occurrences is equal to one more then the current node. We add our word to the
                    # next node. If not, then we create a new node there.
                    if next_node[0] == current_node[0] + 1:
                        next_node[1].append(word)
                    else:
                        self.nodes.insert(i + 1, [current_node[0] + 1, [word]])

                # If there are no words left in this node. We delete the node.
                if len(words) < 1:
                    del self.nodes[i]

                return

        # We check if the first nodes occurrences is 1. If it is we add our word. If not, we create a new node that
        # the occurrences is one
        if self.nodes[0][0] == 1:
            self.nodes[0][1].append(word)
        else:
            self.nodes.insert(0, [1, [word]])

    def random_word(self):
        """ Gets a random word from the histogram

        :return:    The random word
        """

        number = random.random()
        word_count = self.word_count
        last_percent = 0
        for node in self.nodes:
            last_percent = frequency = last_percent + node[0] * len(node[1]) / word_count

            # If the number is greater then the percentage we calculated before, then this is the random node we want.
            # Then we get a random word from the node's second index which is the list of words
            if number < frequency:
                return node[1][random.randint(0, len(node[1]) - 1)]


class Histogram:
    """ Histogram is a class to store types and tokens

     We store all our data in a dictionary. Our key is the current word or phrase, and the value is the number of
     occurrences for that word.
     """

    def __init__(self):
        """ Initialize the data """
        self.nodes = {}
        self.word_count = 0

    def __len__(self):
        """ Gets the amount of words in the data set. Not types! """
        return self.word_count

    def __str__(self):
        """ THe string representation of the nodes """
        return str(self.nodes)

    def update_word(self, current_word):
        """ If the word is already in the histogram we will update the number of times the word occurs. If not then
        we insert it into our nodes

        :param current_word:     The word we want update in the data set
        """

        self.word_count += 1

        if current_word in self.nodes:
            self.nodes[current_word] += 1
        else:
            self.nodes[current_word] = 1

    def random_word(self):
        """ Gets a random word from the histogram

        :return:    The random word
        """

        random_range = random.random()
        total_word_count = self.word_count
        last_percent = 0

        # Go through all our words checking if the one is coming up
        for current_word in self.nodes:
            word_occurrences = self.nodes[current_word]

            # We get the frequency of the word
            last_percent = frequency = last_percent + word_occurrences / total_word_count

            # If the number is greater then the percentage we calculated before, then this is the random node we want.
            if random_range < frequency:
                return current_word

    def __repr__(self):
        return "Histogram[" + str(self.nodes) + "]"


if __name__ == "__main__":
    data = ['one', 'fish', 'two', 'fish', 'red', 'fish', 'blue', 'fish']

    gram = Histogram()
    for word in data:
        gram.update_word(word)

    dd = {}

    for i in range(10000):
        letter = gram.random_word()

        if letter in dd:
            dd[letter] += 1
        else:
            dd[letter] = 1

    print(dd)



from .Dictogram import Dictogram
from collections import deque
from .FileParser import FileParser


class RainbowChain:
    """ Basically a Markov Chain that generates sentences. """

    # Max amount of attempts we have to generate a new sentence
    MAX_ITERATION_ATTEMPTS = 100

    def __init__(self, corpus, max_order):
        """ Constructing the Markov Model

        :param corpus:      The file location of the corpus we want to use
        :param order:       The order for the Markov Chain
        """

        self.parser = FileParser(corpus)
        self.dictogram = []
        self.max_order = max_order

        for order in range(1, max_order + 1):
            self.dictogram.append(Dictogram(self.parser.words, order))

    def generate_sentence(self, backward=False, min_length=50, max_length=140):
        """ Generates a sentence by first getting a sentence start then getting a random token following that word. We
        then end the sentence once we get to an end token ([NONE]). If the sentence is too small or too long we try
        generating a sentence gain.

        :return:    Our uniquely generated sentence
        """

        element = None
        generated_sentence = ''

        if backward:
            window = deque(self.dictogram[self.max_order - 1].random_start(True))
            element = self.dictogram[self.max_order - 1].backwards[tuple(window)]
            generated_sentence = ' '.join(list(reversed(window)))
        else:
            window = deque(self.dictogram[self.max_order - 1].random_start(False))
            element = self.dictogram[self.max_order - 1].forwards[tuple(window)]
            generated_sentence = ' '.join(window)

        # If the window has a split
        if '[SPLIT]' in generated_sentence:
            return self.generate_sentence(backward)

        # We could use a while loop, but we do this instead because we want to make sure we never get an infinite loop
        for _ in range(self.MAX_ITERATION_ATTEMPTS):
            word = None

            # We get a new word or phrase
            current_word = element.random_word()

            # If the word is a sentence end, we finish off the sentence.
            if current_word == '[SPLIT]':
                if not backward:
                    generated_sentence += '.'

                break

            # We make the new word or phrase our current element
            window.popleft()
            window.append(current_word)

            if backward:
                element = self.dictogram[self.max_order - 1].backwards[tuple(window)]

                # We only use the first word in the phrase
                word = current_word + " "

                # Add current word to our new sentence
                generated_sentence = word + generated_sentence
            else:
                element = self.dictogram[self.max_order - 1].forwards[tuple(window)]

                word = " " + current_word

                generated_sentence += word

        sentence_length = len(generated_sentence)

        # If our sentence is too short or long we return a new sentence
        if sentence_length > max_length or sentence_length < min_length:
            return self.generate_sentence(backward)

        # If we are backwards add a period at the end of a sentence
        if backward:
            generated_sentence += '.'

        # Return our sentence and capitalize it. Also make sure there are no uncalled for None tokens
        return generated_sentence

    def generate_with_seed(self, raw_seed, min_length=50, max_length=140, depth=0):
        seed = list(raw_seed.split())

        while len(seed) < self.max_order:
            try:
                word = self.dictogram[len(seed) - 1].forwards[tuple(seed)].random_word()

                if word == '[SPLIT]':
                    return self.generate_with_seed(raw_seed)

                seed.append(word)
            except KeyError:
                return None

        seed = tuple(seed)

        if seed in self.dictogram[self.max_order - 1].forwards and tuple(reversed(seed)) in self.dictogram[
                    self.max_order - 1].backwards:
            forward_element = self.dictogram[self.max_order - 1].forwards[seed]
            forward_window = deque(seed)
            forward_sentence = ''

            for _ in range(self.MAX_ITERATION_ATTEMPTS):
                word = None

                # We get a new word or phrase
                current_word = forward_element.random_word()

                # If the word is a sentence end, we finish off the sentence.
                if current_word == '[SPLIT]':
                    forward_sentence += '.'
                    break

                # We make the new word or phrase our current element
                forward_window.popleft()
                forward_window.append(current_word)

                forward_element = self.dictogram[self.max_order - 1].forwards[tuple(forward_window)]
                word = " " + current_word
                forward_sentence += word

            forward_length = len(forward_sentence)

            # If our sentence is too short or long we return a new sentence
            if forward_length > max_length / 2 or forward_length < min_length / 2:
                if depth > self.MAX_ITERATION_ATTEMPTS:
                    return None

                return self.generate_with_seed(raw_seed, depth=depth + 1)

            backward_window = deque(tuple(reversed(seed)))
            backward_element = self.dictogram[self.max_order - 1].backwards[tuple(reversed(seed))]
            backward_sentence = ' '.join(list(reversed(backward_window)))

            # If the window has a split
            if '[SPLIT]' in backward_sentence:
                if depth > self.MAX_ITERATION_ATTEMPTS:
                    return None

                return self.generate_with_seed(raw_seed, depth=depth + 1)

            for _ in range(self.MAX_ITERATION_ATTEMPTS):
                word = None

                # We get a new word or phrase
                current_word = backward_element.random_word()

                # If the word is a sentence end, we finish off the sentence.
                if current_word == '[SPLIT]':
                    break

                # We make the new word or phrase our current element
                backward_window.popleft()
                backward_window.append(current_word)

                backward_element = self.dictogram[self.max_order - 1].backwards[tuple(backward_window)]
                word = current_word + " "
                backward_sentence = word + backward_sentence

            backward_length = len(backward_sentence)

            # If our sentence is too short or long we return a new sentence
            if backward_length > max_length / 2 or backward_length < min_length / 2:
                if depth > self.MAX_ITERATION_ATTEMPTS:
                    return None

                return self.generate_with_seed(raw_seed, depth=depth + 1)

            return backward_sentence + forward_sentence



from distutils.core import setup

setup(
    name='RainbowChain',
    packages=['RainbowChain'],
    license="MIT License",
    version='0.1.4',
    description='Implementing Middle Out (Indexing) to the Markov Chain',
    author='Avery Durrant',
    author_email='avery@frostbyte.co',
    url='https://github.com/Avery246813579/Python-Rainbow-Chain',
    download_url='https://github.com/Avery246813579/Python-Rainbow-Chain/0.1.4',
    keywords=['Markov Chain', 'Middle Out', 'Markov Model'],
    classifiers=[]
)



import random
import pickle

from .Histogram import Histogram
from collections import deque


class Dictogram:
    """ A Dictogram is a custom data type we use to store data for our Markov Chain. We use key (a word or phrase) then
    a histogram (of following words or phrases) to store our data.

    I refer to key and value a lot during this class. A key is the current word or phrase we are looking at (it can be
    a word of phrase because of the order we are using). A value is the word or phrase following our current key.
     """

    def __init__(self, words, order):
        """ Constructing the Dictogram

        :param words:       The words you want to add (the corpus)
        :param order:       The Markov Chain order
        """

        self.forwards = dict()
        self.backwards = dict()
        self.word_count = len(words)

        self.data = []
        self.sentence_starts = []
        self.sentence_ends = []

        self._construct_data(self.forwards, order, words, False)
        self._construct_data(self.backwards, order, list(reversed(words)), True)

    def _construct_data(self, data, order, words, backward):
        creating = False

        words_length = len(words)
        window = deque(words[0: order])

        if backward and window[0][-1] == '.':
            window[0] = window[0][:-1]

        # Used to find the start of sentences
        if backward:
            self.sentence_ends.append(tuple(window))
        else:
            self.sentence_starts.append(tuple(window))

        # We loop through all our words and if it has occurred we add the following word to a histogram. If it has not
        # occurred we construct a new histogram
        for i in range(order, words_length):
            current_word = words[i]

            if len(window) < order:
                window.append(current_word)
                continue

            if creating:
                if backward:
                    self.sentence_ends.append(tuple(window))
                else:
                    self.sentence_starts.append(tuple(window))

                creating = False

            if self._next_item(data, backward, window, current_word):
                creating = True
                window.clear()

                if backward:
                    window.appendleft(current_word[:-1])

    def _next_item(self, data, backward, window, word):
        split = False

        if word[-1] == '.':
            if backward:
                self._next_item(data, backward, window, '[SPLIT]')

                return True
            else:
                word = word[:-1]
                split = True

        # Add current data to our Dictogram
        self.add(data, tuple(window), word)

        # Add the next number in the sequence
        window.append(word)

        if backward:
            window.popleft()
        else:
            if window.popleft() == '[SPLIT]':
                data.append(tuple(window))

        # End of Word
        if split:
            self._next_item(data, backward, window, '[SPLIT]')

            return True

    def random_start(self, backwards=False):
        """ Finds a random start to our sentence using the end keys list.

        :return:        A key to use in order to construct the start of a sentence
        """
        if backwards:
            return self.sentence_ends[random.randint(0, len(self.sentence_ends) - 1)]
        else:
            return self.sentence_starts[random.randint(0, len(self.sentence_starts) - 1)]

    @staticmethod
    def add(data, key, value):
        """ Adds a key-value pair to the data set.

         :param key:        The word or phrase that we are currently looking at
         :param value:      The value is the word or phrase following the key
         """

        # If the word or phrase has not been said before, we create a new histogram for that word.
        if key not in data:
            data[key] = Histogram()

        # Update the word phrase in the keys histogram
        data[key].update_word(value)

    def random_forward_key(self):
        """ Gets a random word or phrase from our data set """
        keys = list(self.forwards.keys())

        return keys[random.randint(0, len(keys) - 1)]

    def random_backward_key(self):
        """ Gets a random word or phrase from our data set """
        keys = list(self.backwards.keys())

        return keys[random.randint(0, len(keys) - 1)]

    def __str__(self):
        return str(self.forwards) + '\n' + str(self.backwards)